import hashlib
import hmac
import json
import logging
from datetime import datetime

from werkzeug.exceptions import Forbidden

from odoo import http
from odoo.http import request
from odoo.tools import consteq

from odoo.addons.payment_stripe import utils as stripe_utils
from odoo.addons.payment_stripe.controllers.main import StripeController as BaseController
from odoo.addons.saas_worker.util import version_gte


_logger = logging.getLogger(__name__)


class StripeController(BaseController):

    @http.route()
    def stripe_webhook(self):
        """ Override of payment_stripe to process `account.updated` events.

        :return: An empty string to acknowledge the notification with an HTTP 200 response
        :rtype: str
        """
        res = super().stripe_webhook()

        # Compatible with both the old HTTP API and with the HTTP apocalypse of 15.3
        event = json.loads(request.httprequest.data)

        if event['type'] == 'account.updated':
            # Check the source and integrity of the event
            if version_gte('16.0'):
                PaymentProvider = request.env['payment.provider']
            else:
                PaymentProvider = request.env['payment.acquirer']
            provider_sudo = PaymentProvider.sudo().search(
                [('stripe_account_id', '=', event['data']['object']['id'])], limit=1
            )
            if provider_sudo:
                webhook_secret = stripe_utils.get_webhook_secret(provider_sudo)
                self._verify_account_notification_signature(webhook_secret)

                # Update the onboarding status
                if provider_sudo._stripe_onboarding_is_ongoing():
                    provider_sudo._update_stripe_onboarding_status(
                        connected_account=event['data']['object']
                    )
        return res

    def _verify_account_notification_signature(self, webhook_secret):
        """ Check that the received signature matches the expected one.

        This method is only used for events specific to the Stripe 'account' object as the method
        used to check the signature in payment_stripe takes a payment.transaction as argument.
        """
        if not webhook_secret:
            _logger.warning("ignored webhook event due to undefined webhook secret")
            return

        notification_payload = request.httprequest.data.decode('utf-8')
        signature_entries = request.httprequest.headers['Stripe-Signature'].split(',')
        signature_data = {k: v for k, v in [entry.split('=') for entry in signature_entries]}

        # Retrieve the timestamp from the data
        event_timestamp = int(signature_data.get('t', '0'))
        if not event_timestamp:
            _logger.warning("received notification with missing timestamp")
            raise Forbidden()

        # Check if the timestamp is not too old
        if datetime.utcnow().timestamp() - event_timestamp > self.WEBHOOK_AGE_TOLERANCE:
            _logger.warning("received notification with outdated timestamp: %s", event_timestamp)
            raise Forbidden()

        # Retrieve the received signature from the data
        received_signature = signature_data.get('v1')
        if not received_signature:
            _logger.warning("received notification with missing signature")
            raise Forbidden()

        # Compare the received signature with the expected signature computed from the data
        signed_payload = f'{event_timestamp}.{notification_payload}'
        expected_signature = hmac.new(
            webhook_secret.encode('utf-8'), signed_payload.encode('utf-8'), hashlib.sha256
        ).hexdigest()
        if not consteq(received_signature, expected_signature):
            _logger.warning("received notification with invalid signature")
            raise Forbidden()
